# -*- coding: utf-8 -*-
"""
This module defines general classes to describes wakefields, impedances and 
wake functions. Based on impedance library by David Amorim.

@author: David Amorin, Alexis Gamelin
@date: 14/01/2020
"""

import warnings
import re
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import collective_effects.tools as tools
from scipy.interpolate import interp1d
from tracking.element import Element
from copy import deepcopy
from scipy.integrate import trapz

class ComplexData:
    """
    Define a general data structure for a complex function based on a pandas 
    DataFrame.

    Parameters
    ----------
    variable : list, numpy array
       contains the function variable values

    function : list or numpy array of complex numbers
        contains the values taken by the complex function
    """

    def __init__(self, variable=np.array([-1e15, 1e15]),
                 function=np.array([0, 0])):
        self.data = pd.DataFrame({'real': np.real(function),
                                  'imag': np.imag(function)},
                                 index=variable)
        self.data.index.name = 'variable'

    def add(self, structure_to_add, method='zero', interp_kind='cubic', 
            index_name="variable"):
        """
        Method to add two structures. If the data don't have the same length,
        different cases are handled.
        
        Parameters
        ----------
        structure_to_add : ComplexData object, int, float or complex.
            structure to be added.
        method : str, optional
            manage how the addition is done, possibilties are: 
            -common: the common range of the index (variable) from the two
                structures is used. The input data are cross-interpolated
                so that the result data contains the points at all initial
                points in the common domain.
            -extrapolate: extrapolate the value of both ComplexData.
            -zero: outside of the common range, the missing values are zero. In 
                the common range, behave as "common" method.
        interp_kind : str, optional
            interpolation method which is passed to pandas and to
            scipy.interpolate.interp1d.
        index_name : str, optional
            name of the dataframe index passed to the method
            
        Returns
        -------
        ComplexData 
            Contains the sum of the two inputs.
        """

        # Create first two new DataFrames merging the variable
        # from the two input data

        if isinstance(structure_to_add, (int, float, complex)):
            structure_to_add = ComplexData(variable=self.data.index,
                                           function=(structure_to_add * 
                                                     np.ones(len(self.data.index))))
                                
        data_to_concat = structure_to_add.data.index.to_frame().set_index(index_name)
        
        initial_data = pd.concat([self.data, data_to_concat], sort=True)
        initial_data = initial_data[~initial_data.index.duplicated(keep='first')]
        initial_data = initial_data.sort_index()

        data_to_add = pd.concat(
                        [structure_to_add.data,
                         self.data.index.to_frame().set_index(index_name)],
                        sort=True)
        data_to_add = data_to_add[~data_to_add.index.duplicated(keep='first')]
        data_to_add = data_to_add.sort_index()

        if method == 'common':
            max_variable = min(structure_to_add.data.index.max(),
                               self.data.index.max())

            min_variable = max(structure_to_add.data.index.min(),
                               self.data.index.min())

            initial_data = initial_data.interpolate(method=interp_kind)
            mask = ((initial_data.index <= max_variable)
                    & (initial_data.index >= min_variable))
            initial_data = initial_data[mask]

            data_to_add = data_to_add.interpolate(method=interp_kind)
            mask = ((data_to_add.index <= max_variable)
                    & (data_to_add.index >= min_variable))
            data_to_add = data_to_add[mask]

            result_structure = ComplexData()
            result_structure.data = initial_data + data_to_add
            return result_structure

        if method == 'extrapolate':
            print('Not there yet')
            return self
        
        if method == 'zero':
            max_variable = min(structure_to_add.data.index.max(),
                               self.data.index.max())

            min_variable = max(structure_to_add.data.index.min(),
                               self.data.index.min())

            mask = ((initial_data.index <= max_variable)
                    & (initial_data.index >= min_variable))
            initial_data[mask] = initial_data[mask].interpolate(method=interp_kind)

            mask = ((data_to_add.index <= max_variable)
                    & (data_to_add.index >= min_variable))
            data_to_add[mask] = data_to_add[mask].interpolate(method=interp_kind)
            
            initial_data.replace(to_replace=np.nan, value=0, inplace=True)
            data_to_add.replace(to_replace=np.nan, value=0, inplace=True)
            
            result_structure = ComplexData()
            result_structure.data = initial_data + data_to_add
            return result_structure

    def __radd__(self, structure_to_add):
        return self.add(structure_to_add, method='zero')

    def __add__(self, structure_to_add):
        return self.add(structure_to_add, method='zero')

    def multiply(self, factor):
        """
        Multiply a data strucure with a float or an int.
        If the multiplication is done with something else, throw a warning.
        """
        if isinstance(factor, (int, float)):
            result_structure = ComplexData()
            result_structure.data = self.data * factor
            return result_structure
        else:
            warnings.warn(('The multiplication factor is not a float '
                           'or an int.'), UserWarning)
            return self

    def __mul__(self, factor):
        return self.multiply(factor)

    def __rmul__(self, factor):
        return self.multiply(factor)
    
    def __call__(self, values, interp_kind="cubic"):
        """
        Interpolation of the data by calling the class to have a function-like
        behaviour.
        
        Parameters
        ----------
        values : list or numpy array of complex, int or float
            values to be interpolated.
        interp_kind : str, optional
            interpolation method which is passed to scipy.interpolate.interp1d.
            
        Returns
        -------
        numpy array 
            Contains the interpolated data.
        """
        real_func = interp1d(x = self.data.index, 
                             y = self.data["real"], kind=interp_kind)
        imag_func = interp1d(x = self.data.index, 
                             y = self.data["imag"], kind=interp_kind)
        return real_func(values) + 1j*imag_func(values)
    
class WakeFunction(ComplexData):
    """
    Define a WakeFunction object based on a ComplexData object.
    """

    def __init__(self,
                 variable=np.array([-1e15, 1e15]),
                 function=np.array([0, 0]), wake_type='long'):
        super().__init__(variable, function)
        pass
    
class Impedance(ComplexData):
    """
    Define an Impedance object based on a ComplexData object.
    """

    def __init__(self,
                 variable=np.array([-1e15, 1e15]),
                 function=np.array([0, 0]), impedance_type='long'):
        super().__init__(variable, function)
        self._impedance_type = impedance_type
        self.data.index.name = 'frequency [Hz]'
        self.initialize_impedance_coefficient()


    def initialize_impedance_coefficient(self):
        """
        Define the impedance coefficients and the plane of the impedance that
        are presents in attributes of the class.
        """
        table = self.impedance_name_and_coefficients_table()
        
        try:
            component_coefficients = table[self.impedance_type].to_dict()
        except KeyError:
            print('Impedance type {} does not exist'.format(self.impedance_type))
        
        self.a = component_coefficients["a"]
        self.b = component_coefficients["b"]
        self.c = component_coefficients["c"]
        self.d = component_coefficients["d"]
        self.plane = component_coefficients["plane"]
                    
    def impedance_name_and_coefficients_table(self):
        """
        Return a table associating the human readbale names of an impedance
        component and its associated coefficients and plane.
        """

        component_dict = {
            'long': {'a': 0, 'b': 0, 'c': 0, 'd': 0, 'plane': 'z'},
            'xcst': {'a': 0, 'b': 0, 'c': 0, 'd': 0, 'plane': 'x'},
            'ycst': {'a': 0, 'b': 0, 'c': 0, 'd': 0, 'plane': 'y'},
            'xdip': {'a': 1, 'b': 0, 'c': 0, 'd': 0, 'plane': 'x'},
            'ydip': {'a': 0, 'b': 1, 'c': 0, 'd': 0, 'plane': 'y'},
            'xydip': {'a': 0, 'b': 1, 'c': 0, 'd': 0, 'plane': 'x'},
            'yxdip': {'a': 1, 'b': 0, 'c': 0, 'd': 0, 'plane': 'y'},
            'xquad': {'a': 0, 'b': 0, 'c': 1, 'd': 0, 'plane': 'x'},
            'yquad': {'a': 0, 'b': 0, 'c': 0, 'd': 1, 'plane': 'y'},
            'xyquad': {'a': 0, 'b': 0, 'c': 0, 'd': 1, 'plane': 'x'},
            'yxquad': {'a': 0, 'b': 0, 'c': 1, 'd': 0, 'plane': 'y'},
            }

        return pd.DataFrame(component_dict)

    def add(self, structure_to_add, beta_x=1, beta_y=1, method='zero'):
        """
        Method to add two Impedance objects. The two structures are
        compared so that the addition is self-consistent.
        Beta functions can be precised as well.
        """

        if isinstance(structure_to_add, (int, float, complex)):
            result = super().add(structure_to_add, method=method, 
                                 index_name="frequency [Hz]")
        elif isinstance(structure_to_add, Impedance):
            if (self.impedance_type == structure_to_add.impedance_type):
                weight = (beta_x ** self.power_x) * (beta_y ** self.power_y)
                result = super().add(structure_to_add * weight, method=method, 
                                     index_name="frequency [Hz]")
            else:
                warnings.warn(('The two Impedance objects do not have the '
                               'same coordinates or plane or type. '
                               'Returning initial Impedance object.'),
                              UserWarning)
                result = self
                
        impedance_to_return = Impedance(
                                result.data.index,
                                result.data.real.values + 1j*result.data.imag.values,
                                self.impedance_type)    
        return impedance_to_return

    def __radd__(self, structure_to_add):
        return self.add(structure_to_add)

    def __add__(self, structure_to_add):
        return self.add(structure_to_add)

    def multiply(self, factor):
        """
        Multiply a Impedance object with a float or an int.
        If the multiplication is done with something else, throw a warning.
        """
        result = super().multiply(factor)
        impedance_to_return = Impedance(
                                result.data.index,
                                result.data.real.values + 1j*result.data.imag.values,
                                self.impedance_type)    
        return impedance_to_return

    def __mul__(self, factor):
        return self.multiply(factor)

    def __rmul__(self, factor):
        return self.multiply(factor)
        
    @property
    def impedance_type(self):
        return self._impedance_type
    
    @impedance_type.setter
    def impedance_type(self, value):
        self._impedance_type = value
        self.initialize_impedance_coefficient()

    @property
    def power_x(self):
        power_x = self.a/2 + self.c/2.
        if self.plane == 'x':
            power_x += 1./2.
        return power_x

    @property
    def power_y(self):
        power_y = self.b/2. + self.d/2.
        if self.plane == 'y':
            power_y += 1./2.
        return power_y
    
class Wakefield:
    """
    Defines a Wakefield which corresponds to a single physical element which 
    produces different types of wakes, represented by their Impedance or 
    WakeFunction objects.
    
    Parameters
    ----------
    structure_list : list, optional
        list of Impedance/WakeFunction objects to add to the Wakefield.
    name : str, optional
    
    Attributes
    ----------
    impedance_components : array of str
        Impedance component names present for this element.
        
    Methods
    -------
    append_to_model(structure_to_add)
        Add Impedance/WakeFunction to Wakefield.
    list_to_attr(structure_list)
        Add list of Impedance/WakeFunction to Wakefield.
    add_wakefields(wake1, beta1, wake2, beta2)
        Add two Wakefields taking into account beta functions.
    add_several_wakefields(wakefields, beta)
        Add a list of Wakefields taking into account beta functions.
    """

    def __init__(self, structure_list=None, name=None):
        self.list_to_attr(structure_list)
        self.name = name

    def append_to_model(self, structure_to_add):
        """
        Add Impedance/WakeFunction component to Wakefield.

        Parameters
        ----------
        structure_to_add : Impedance or WakeFunction object
        """
        list_of_attributes = dir(self)
        if isinstance(structure_to_add, Impedance):
            attribute_name = "Z" + structure_to_add.impedance_type
            if attribute_name in list_of_attributes:
                raise ValueError("There is already a component of the type "
                                 "{} in this element.".format(attribute_name))
            else:
                self.__setattr__(attribute_name, structure_to_add)
        elif isinstance(structure_to_add, WakeFunction):
            attribute_name = "W" + structure_to_add.impedance_type
            if attribute_name in list_of_attributes:
                raise ValueError("There is already a component of the type "
                                 "{} in this element.".format(attribute_name))
            else:
                self.__setattr__(attribute_name, structure_to_add)
        else:
            raise ValueError("{} is not an Impedance nor a WakeFunction.".format(structure_to_add))
    
    def list_to_attr(self, structure_list):
        """
         Add list of Impedance/WakeFunction components to Wakefield.

        Parameters
        ----------
        structure_list : list of Impedance or WakeFunction objects.
        """
        if structure_list is not None:
            for component in structure_list:
                self.append_to_model(component)
    
    @property
    def impedance_components(self):
        """
        Return an array of the impedance component names for the element.
        """
        return np.array([comp for comp in dir(self) if re.match(r'[Z]', comp)])
    
    @staticmethod
    def add_wakefields(wake1, beta1, wake2, beta2):
        """
        Add two Wakefields taking into account beta functions.

        Parameters
        ----------
        wake1 : Wakefield
            Wakefield to add.
        beta1 : array of shape (2,)
            Beta functions at wake1 postion.
        wake2 : Wakefield
            Wakefield to add.
        beta2 : array of shape (2,)
            Beta functions at wake2 postion.

        Returns
        -------
        wake_sum : Wakefield
            Wakefield with summed components.

        """
        wake_sum = deepcopy(wake1)
        for component_name1 in wake1.impedance_components:
            impedance1 = getattr(wake_sum, component_name1)
            weight1 = ((beta1[0] ** impedance1.power_x) * 
                      (beta1[1] ** impedance1.power_y))
            setattr(wake_sum, component_name1, weight1*impedance1)
            
        for component_name2 in wake2.impedance_components: 
            impedance2 = getattr(wake2, component_name2)
            weight2 = ((beta2[0] ** impedance2.power_x) * 
                      (beta2[1] ** impedance2.power_y))
            try:
                impedance1 = getattr(wake_sum, component_name2)
                setattr(wake_sum, component_name2, impedance1 +
                        weight2*impedance2)
            except AttributeError:
                setattr(wake_sum, component_name2, weight2*impedance2)

        return wake_sum
    
    @staticmethod
    def add_several_wakefields(wakefields, beta):
        """
        Add a list of Wakefields taking into account beta functions.
        
        Parameters
        ----------
        wakefields : list of Wakefield
            Wakefields to add.
        beta : array of shape (len(wakefields), 2)
            Beta functions at the Wakefield postions.

        Returns
        -------
        wake_sum : Wakefield
            Wakefield with summed components..

        """
        if len(wakefields) == 1:
            return wakefields[0]
        elif len(wakefields) > 1:
            wake_sum = Wakefield.add_wakefields(wakefields[0], beta[:,0],
                                     wakefields[1], beta[:,1])
            for i in range(len(wakefields) - 2):
                wake_sum = Wakefield.add_wakefields(wake_sum, [1 ,1], 
                                         wakefields[i+2], beta[:,i+2])
            return wake_sum
        else:
            raise ValueError("Error in input.")
        
    
class ImpedanceModel(Element):
    """
    Define the impedance model of the machine.
    
    Parameters
    ----------
    ring : Synchrotron object
    wakefield_list : list of Wakefield objects
        Wakefields to add to the model.
    wakefiled_postions : list
        Longitudinal positions corresponding to the added Wakfields.
    
    Attributes
    ----------
    wakefields : list of Wakefield objects
        Wakefields in the model.
    positions : array
        Wakefields positions.
    names : array
        Names of (unique) Wakefield objects.
    sum : Wakefield
        Sum of every Wakefield in the model.
    sum_"name" : Wakefield
        Sum of every Wakefield with the same "name".
    sum_names : array
        Names of attributes where the wakefields are summed by name.    
    
    Methods
    -------
    add(wakefield_list, wakefiled_postions)
        Add elements to the model.
    add_multiple_elements(wakefield, wakefiled_postions)
        Add the same element at different locations to the model.
    sum_elements()
        Sum all wakefields into self.sum.
    update_name_list()
        Update self.names with uniques names of self.wakefields.
    find_wakefield(name)
        Return indexes of wakefields with the same name in self.wakefields.
    sum_by_name(name)
        Sum the elements with the same name in the model into sum_name.
    sum_by_name_all(name)
        Sum all the elements with the same name in the model into sum_name.
    plot_area(Z_type="Zlong", component="real", sigma=None, attr_list=None)
        Plot the contributions of different kind of wakefields.
    """
    
    def __init__(self, ring, wakefield_list=None, wakefiled_postions=None):
        self.ring = ring
        self.optics = self.ring.optics
        self.wakefields = []
        self.positions = np.array([])
        self.names = np.array([])
        self.sum_names = np.array([])
        self.add(wakefield_list, wakefiled_postions)
        
    def track(self, beam):
        """
        Track a beam object through this Element.
        
        Parameters
        ----------
        beam : Beam object
        """
        raise NotImplementedError
        
    def sum_elements(self):
        """Sum all wakefields into self.sum"""
        beta = self.optics.beta(self.positions)
        self.sum = Wakefield.add_several_wakefields(self.wakefields, beta)
        if "sum" not in self.sum_names:
            self.sum_names = np.append(self.sum_names, "sum")
    
    def update_name_list(self):
        """Update self.names with uniques names of self.wakefields."""
        for wakefield in self.wakefields:
            if wakefield.name is None:
                continue
            if wakefield.name not in self.names:
                self.names = np.append(self.names, wakefield.name)
                
    def count_elements(self):
        """Count number of each type of wakefield in the model."""
        self.count = np.zeros(len(self.names))
        for wakefield in self.wakefields:
            if wakefield.name is not None:
                self.count += (wakefield.name == self.names).astype(int)
                
    def find_wakefield(self, name):
        """
        Return indexes of wakefields with the same name in self.wakefields.

        Parameters
        ----------
        name : str
            Wakefield name.

        Returns
        -------
        index : list
            Index of positions in self.wakefields.

        """
        index = []
        for i, wakefield in enumerate(self.wakefields):
            if wakefield.name == name:
                index.append(i)
        return index

    def sum_by_name(self, name):
        """
        Sum the elements with the same name in the model into sum_name.
        
        Parameters
        ----------
        name : str
            Name of the wakefield to sum.
        """
        attribute_name = "sum_" + name
        index = self.find_wakefield(name)
        beta = self.optics.beta(self.positions[index])
        wakes = []
        for i in index:
            wakes.append(self.wakefields[i])
        wake_sum = Wakefield.add_several_wakefields(wakes, beta)
        setattr(self, attribute_name, wake_sum)
        if attribute_name not in self.sum_names:
            self.sum_names = np.append(self.sum_names, attribute_name)
            
    def sum_by_name_all(self):
        """
        Sum all the elements with the same name in the model into sum_name.
        """
        for name in self.names:
            self.sum_by_name(name)
                    
    def add(self, wakefield_list, wakefiled_postions):
        """
        Add elements to the model.

        Parameters
        ----------
        wakefield_list : list of Wakefield objects
            Wakefields to add to the model.
        wakefiled_postions : list
            Longitudinal positions corresponding to the added Wakfields.
        """
        if (wakefield_list is not None) and (wakefiled_postions is not None):
            for wakefield in wakefield_list:
                self.wakefields.append(wakefield)
                
            for position in wakefiled_postions:
                self.positions = np.append(self.positions, position)
                
        self.update_name_list()
                
    def add_multiple_elements(self, wakefield, wakefiled_postions):
        """
        Add the same element at different locations to the model.

        Parameters
        ----------
        wakefield : Wakefield object
            Wakefield to add to the model.
        wakefiled_postions : list
            Longitudinal positions corresponding to the added Wakfield.
        """
        for position in wakefiled_postions:
            self.positions = np.append(self.positions, position)
            self.wakefields.append(wakefield)
            
        self.update_name_list()
            
    def plot_area(self, Z_type="Zlong", component="real", sigma=None, 
                  attr_list=None):
        """
        Plot the contributions of different kind of wakefields.

        Parameters
        ----------
        Z_type : str, optional
            Type of impedance to plot.
        component : str, optional
            Component to plot, can be "real" or "imag". 
        sigma : float, optional
            RMS bunch length in [s] to use for the spectral density. If equal
            to None, the spectral density is not plotted.
        attr_list : list or array of str, optional
            Attributes to plot. 

        """
        if attr_list is None:
            attr_list = self.sum_names[self.sum_names != "sum"]
        
        # sort plot by decresing area size        
        area = np.zeros((len(attr_list),))
        for index, attr in enumerate(attr_list):
            sum_imp = getattr(getattr(self, attr), Z_type)
            area[index] = trapz(sum_imp.data[component], sum_imp.data.index)
        sorted_index = area.argsort()[::-1]
        
        # Init fig
        fig = plt.figure()
        ax = fig.gca()
        zero_impedance = getattr(self.sum, Z_type)*0
        total_imp = 0
        legend = []
        
        if sigma is not None:
            legend.append("Spectral density for sigma = " + str(sigma) + " s")
        
        # Main plot
        for index in  sorted_index:
            attr = attr_list[index]
            # Set all impedances with common indexes using + zero_impedance
            sum_imp = getattr(getattr(self, attr), Z_type) + zero_impedance
            ax.fill_between(sum_imp.data.index*1e-9, total_imp, 
                            total_imp + sum_imp.data[component])
            total_imp += sum_imp.data[component]
            if attr_list is None:
                legend.append(attr[4:])
            else:
                legend.append(attr)
            
        if sigma is not None:
            spect = tools.spectral_density(zero_impedance.data.index, sigma)
            spect = spect/spect.max()*total_imp.max()
            ax.plot(sum_imp.data.index*1e-9, spect, 'r', linewidth=2.5)
        
        ax.legend(legend, loc="upper left")            
        ax.set_xlabel("Frequency [GHz]")
        
        if Z_type == "Zlong":
            ax.set_ylabel(Z_type + " [Ohm] - " + component + " part")
        else:
            ax.set_ylabel(Z_type + " [Ohm/m] - " + component + " part")
            
        return fig
            
    def summary(self,  sigma, attr_list=None, list_components=None):
        
        if attr_list is None:
            attr_list = self.sum_names
        if list_components is None:
            list_components = self.sum.impedance_components
        
        shape_array = (len(list_components), len(attr_list))
        loss_array = np.zeros(shape_array)
        
        for i, attr in enumerate(attr_list):
            for j, component_name in enumerate(list_components):
                try:
                    impedance = getattr(getattr(self, attr), component_name)
                    loss_array[j,i] = tools.loss_factor(impedance, sigma)
                except AttributeError:
                        pass
                    
        summary = pd.DataFrame(loss_array.T, index=attr_list, 
                               columns=list_components)
        return summary
    
    def energy_loss(self, sigma, M, bunch_spacing, I, n_points=10e6):
        """
        Compute the beam and bunch loss factor and energy losses for each type 
        of element in the model assuming Gaussian bunches and constant spacing
        between bunches.

        Parameters
        ----------
        sigma : float
            RMS bunch length in [s].
        M : int
            Number of bunches in the beam.
        bunch_spacing : float
            Time between two bunches in [s].
        I : float
            Total beam current in [A].
        n_points : float, optional
            Number of points used in the frequency spectrums.

        Returns
        -------
        summary : Dataframe
            Contains the beam and bunch loss factor and energy loss for the 
            full model and for each type of different component.

        """
        
        fmax = self.sum.Zlong.data.index.max()
        fmin = self.sum.Zlong.data.index.min()
        
        Q = I*self.ring.T0/M
        
        if fmin >= 0:
            fmin = -1*fmax
        f = np.linspace(fmin, fmax, int(n_points))

        beam_spect = tools.beam_spectrum(f, M, bunch_spacing, sigma= sigma)
        
        bunch_spect = tools.Gaussian_bunch_spectrum(f, sigma)
        
        attr_list = self.sum_names
        
        loss_array = np.zeros((len(attr_list),2))
        
        for i, attr in enumerate(attr_list):
            try:
                impedance = getattr(getattr(self, attr), "Zlong")
                loss_array[i,0] = tools.beam_loss_factor(impedance, f, beam_spect, self.ring)
                loss_array[i,1] = tools.beam_loss_factor(impedance, f, bunch_spect, self.ring)
            except AttributeError:
                pass

        loss_array = loss_array*1e-12
        summary = pd.DataFrame(loss_array, index=attr_list, 
                               columns=["loss factor (beam) [V/pC]", "loss factor (bunch) [V/pC]"])
        
        summary["P (beam) [W]"] = summary["loss factor (beam) [V/pC]"]*1e12*Q**2/(self.ring.T0)
        summary["P (bunch) [W]"] = summary["loss factor (bunch) [V/pC]"]*1e12*Q**2/(self.ring.T0)*M
                
        return summary
        
        
    